# Spacecraft Motion Simulator by Tetsuya Kusumoto and Tiago Monteiro for Patchedconics, LLC.

# Packages Import
import numpy as np
from numpy import sin, cos
from numpy.random import *
import tkinter as tk
from PIL import ImageTk, Image
import os
import time
import sys

# Simulates the attitude dynamics of the spacecraft
def pnp(cube, omega1, omega2, omega3, Xg, Yg, Zg, phi1, phi2, phi3, vx, vy, vz, dts, dte, B, ST, IL, marker1list, marker2list, marker3list, marker4list, para_error, tlist,f, n, Iall, surf, paths):

    import matplotlib.pyplot as plt
    import subprocess
    import pandas as pd
    import csv

    clearConsole = lambda: os.system('cls' if os.name in ('nt', 'dos') else 'clear')

    export = []
        
    export.append(["t", "Ixx_true", "Iyy_true", "Izz_true", "Ixy_true", "Ixz_true", "Iyz_true", "omega1_true", "omega2_true", "omega3_true", "L1_true", "L2_true", "L3_true",
                        "Ixx_est",  "Iyy_est",  "Izz_est",  "Ixy_est",  "Ixz_est",  "Iyz_est",  "omega1_est",  "omega2_est",  "omega3_est",  "L1_est",  "L2_est",  "L3_est",
                        "Markers_hidden?", "Estimator_restart?", "Marker1_x", "Marker1_y", "Marker1_z", "Marker2_x", "Marker2_y", "Marker2_z", "Marker3_x", "Marker3_y",
                        "Marker3_z", "Marker4_x", "Marker4_y", "Marker4_z", "DCM11", "DCM12", "DCM13", "DCM21", "DCM22", "DCM23", "DCM31", "DCM32", "DCM33"])

    pathcd, pathexe, pathtoest, pathtosim, pathshape, pathres, pathresf, pathnpz, pathdata, pathresfo = paths[0], paths[1], paths[2], paths[3], paths[4], paths[5], paths[6], paths[7], paths[8], paths[9]

    qini1 = cos(phi2) * cos(phi1) * sin(phi3 / 2)
    qini2 = cos(phi2) * sin(phi1) * sin(phi3 / 2)
    qini3 = sin(phi2) * sin(phi3 / 2)
    qini4 = cos(phi3 / 2)

    Rinit = np.array([[qini1 ** 2 - qini2 ** 2 - qini3 ** 2 + qini4 ** 2, 2 * (qini1 * qini2 + qini3 * qini4),
                       2 * (qini3 * qini1 - qini2 * qini4)], \
                      [2 * (qini1 * qini2 - qini3 * qini4), qini2 ** 2 - qini3 ** 2 - qini1 ** 2 + qini4 ** 2,
                       2 * (qini2 * qini3 + qini1 * qini4)], \
                      [2 * (qini3 * qini1 + qini2 * qini4), 2 * (qini2 * qini3 - qini1 * qini4),
                       qini3 ** 2 - qini1 ** 2 - qini2 ** 2 + qini4 ** 2]])

    cube3 = np.delete(cube, 3, 1)
    cubelist = [[[0.0, 0.0, 0.0] for i in range(len(cube))] for i in range(n)]

    for j in range(len(cube)):
        cubelist[0][j] = cube3[j]

    R = [np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]) for i in range(n)]
    q1, q2, q3, q4 = 0.0, 0.0, 0.0, 1.0
    qllist = [np.array([[q1], [q2], [q3], [q4]]) for i in range(n)]
    omega = [np.array([[omega1], [omega2], [omega3]]) for i in range(n)]
    t = []
    Itarlist = []

    T = [np.empty((1, 3)) for i in range(n)]
    RT = [np.array((3, 3)) for i in range(n)]
    O = np.array([[0, 0, 0, 1]])

    Pr = np.array([[f, 0, 0, 0], [0, f, 0, 0], [0, 0, 1, 0]])
    projection = [[0 for i in range(len(cube))] for j in range(n)]
    xyz = [[0 for i in range(len(cube))] for j in range(n)]
    xy = [[0 for i in range(len(cube))] for j in range(n)]
    
    interval = int(1/dts)
    timelist, esttimelist = [], []
    omseikai1list, omseikai2list, omseikai3list = [], [], []
    om1list, om2list, om3list = [], [], []
    Lseikai1list, Lseikai2list, Lseikai3list = [], [], []
    L1list, L2list, L3list = [], [], []
    I1estlist, I2estlist, I3estlist, I12estlist, I13estlist, I23estlist = [], [], [], [], [], []
    I1truelist, I2truelist, I3truelist, I12truelist, I13truelist, I23truelist = [], [], [], [], [], []
    fig = plt.figure(figsize=(16, 8))
    ax1 = fig.add_subplot(221)
    ax1.set_xlim([0, n*dts])
    ax2 = fig.add_subplot(222)
    ax2.set_xlim([0, n*dts])
    ax3 = fig.add_subplot(223)
    ax3.set_xlim([0, n*dts])
    ax3.grid(color='grey', alpha=0.3)


    
    steplist = []
    visible_previous = 0
    estcounter = 0
    # -------------------------4次ルンゲクッタ法により積分(コード内ではk1~k4, kq1~kq4が各接線を表す)-----------------------------------
    # Itar: ターゲットの慣性テンソル(I target)¥
    for i in range(n):

        reset = 0
        
        if i == 0:
            counter = dte/dts
        
        currstep = counter*dts
        
        if i < n-1:
            Itar = Iall
            Itarlist.append(Itar)
            k1 = -np.dot(np.linalg.inv(Itar), np.cross(omega[i].T, np.dot(Itar, omega[i]).T).T)
            k2 = -np.dot(np.linalg.inv(Itar),
                         np.cross((omega[i] + k1 / 2 * dts).T, np.dot(Itar, (omega[i] + k1 / 2 * dts)).T).T)
            k3 = -np.dot(np.linalg.inv(Itar),
                         np.cross((omega[i] + k2 / 2 * dts).T, np.dot(Itar, (omega[i] + k2 / 2 * dts)).T).T)
            k4 = -np.dot(np.linalg.inv(Itar),
                         np.cross((omega[i] + k3 * dts).T, np.dot(Itar, (omega[i] + k3 * dts)).T).T)
            omega[i + 1] = omega[i] + (k1 + 2 * k2 + 2 * k3 + k4) / 6 * dts

            o1, o2, o3 = omega[i + 1][0][0], omega[i + 1][1][0], omega[i + 1][2][0]
            Omeskew = np.array([[0, o3, -o2, o1],
                                [-o3, 0, o1, o2],
                                [o2, -o1, 0, o3],
                                [-o1, -o2, -o3, 0]])
            kq1 = 1 / 2 * np.dot(Omeskew, qllist[i])
            kq2 = 1 / 2 * np.dot(Omeskew, (qllist[i] + kq1 / 2 * dts))
            kq3 = 1 / 2 * np.dot(Omeskew, (qllist[i] + kq2 / 2 * dts))
            kq4 = 1 / 2 * np.dot(Omeskew, (qllist[i] + kq3 * dts))

            qllist[i + 1] = qllist[i] + ((kq1 + 2 * kq2 + 2 * kq3 + kq4)/6) * dts
            qllist[i + 1] = qllist[i + 1] / np.linalg.norm(qllist[i + 1])

            t.append(dts * i)
            q1, q2, q3, q4 = qllist[i + 1][0][0], qllist[i + 1][1][0], qllist[i + 1][2][0], qllist[i + 1][3][0]
            R[i + 1] = np.array([[q1 ** 2 - q2 ** 2 - q3 ** 2 + q4 ** 2, 2 * (q1 * q2 + q3 * q4), 2 * (q3 * q1 - q2 * q4)], \
                                 [2 * (q1 * q2 - q3 * q4), q2 ** 2 - q3 ** 2 - q1 ** 2 + q4 ** 2, 2 * (q2 * q3 + q1 * q4)], \
                                 [2 * (q3 * q1 + q2 * q4), 2 * (q2 * q3 - q1 * q4),
                                  q3 ** 2 - q1 ** 2 - q2 ** 2 + q4 ** 2]]).T
            tlist.append(i * dts)

            for j in range(len(cube)):
                if i == 0:
                    cubelist[i][j] = np.dot(R[i], cube3[j].T)
                    cubelist[i][j] = np.dot(Rinit, cubelist[i][j].T)
                cubelist[i + 1][j] = np.dot(R[i + 1], cube3[j].T)
                cubelist[i + 1][j] = np.dot(Rinit, cubelist[i + 1][j].T)
        T[i] = np.array([[Xg + vx * i * dts], [Yg + vy * i * dts], [Zg + vz * i * dts]])
        Rre = np.dot(Rinit, R[i])
        RT[i] = np.c_[Rre, T[i]]
        RT[i] = np.r_[RT[i], O]

        for j in range(len(cube)):
            xy[i][j] = np.dot(RT[i], [[cube[j, 0]], [cube[j, 1]], [cube[j, 2]], [cube[j, 3]]])
            xyz[i][j] = np.dot(Pr, xy[i][j])
            projection[i][j] = [xyz[i][j][0] / xyz[i][j][2], xyz[i][j][1] / xyz[i][j][2]]
            projection[i][j] = [projection[i][j][0][0] + randn() * para_error,
                                projection[i][j][1][0] + randn() * para_error]

        transl = np.array([Xg + vx * i * dts, Yg + vy * i * dts, Zg + vz * i * dts])
        marker1list.append(cubelist[i][28] + transl);
        marker2list.append(cubelist[i][29] + transl);
        marker3list.append(cubelist[i][30] + transl);
        marker4list.append(cubelist[i][31] + transl);
        R[i] = np.dot(Rinit, R[i])
        R[i] = R[i].astype(np.float64)
        n1 = cubelist[i][32]
        n2 = cubelist[i][33]

        n1ang = np.dot(marker2list[i], n1) / np.linalg.norm(marker2list[i]) / np.linalg.norm(n1)
        n2ang = np.dot(marker4list[i], n2) / np.linalg.norm(marker4list[i]) / np.linalg.norm(n2)

        if i == 0:
            
            reset = 1
            
            if os.path.exists(pathtoest):
                os.remove(pathtoest)
            if os.path.exists(pathtosim):
                os.remove(pathtosim)
            if os.path.exists(pathres):
                os.remove(pathres)
        if currstep >= dte-0.000001:
        
            counter = 0
            
            if i == 0:
                counter = 0
            
            visible_flag = 0
            if n1ang >= 0 or n2ang >= 0:
                visible_flag = 1
                projection[i][28][0] = 2000

            input1 = [projection[i][28][0], projection[i][28][1], projection[i][29][0], projection[i][29][1], projection[i][30][0], projection[i][30][1], projection[i][31][0], projection[i][31][1], IL, ST, t[i], dte, B, reset]

            resultlist = np.array([input1])

            #print('\n\n')
            #print("Input sent to estimator is:",input1)
            # print(np.array([partial[28][0],partial[28][1],partial[29][0],partial[29][1],partial[30][0],partial[30][1],partial[31][0],partial[31][1]]))

            np.savetxt(pathtoest, resultlist, delimiter=',')

            subprocess.run([pathexe, pathtoest, pathcd, pathtosim, pathshape])

            if reset == 0:
                steplist.append(i)
                outputfile = pathtosim
                datas = []
                with open(outputfile) as f:
                    reader = csv.reader(f)
                    for row in reader:
                        datas.append(row)
                for k in range(len(datas)):
                    for j in range(len(datas[0])):
                        datas[k][j] = float(datas[k][j])
                Lest = np.array(datas[0])
                omest = np.array(datas[1])
                Iest = np.array([datas[2], datas[3], datas[4]])
                Rest  = np.array([datas[5],datas[6],datas[7],datas[8]])
                DCM   = np.array([datas[9],datas[10],datas[11]])
                timelist.append(t[i])
                Lanswer = np.dot(R[i],np.dot(Itar,omega[i]))
                Lansnorm = Lanswer.T/np.linalg.norm(Lanswer)
                Lseikai1list.append(Lansnorm[0][0])
                Lseikai2list.append(Lansnorm[0][1])
                Lseikai3list.append(Lansnorm[0][2])
                omseikai = omega[i]
                omseikai1list.append(omseikai[0][0])
                omseikai2list.append(omseikai[1][0])
                omseikai3list.append(omseikai[2][0])
                I1truelist.append(Itar[0][0] / np.linalg.norm(Lanswer))
                I2truelist.append(Itar[1][1] / np.linalg.norm(Lanswer))
                I3truelist.append(Itar[2][2] / np.linalg.norm(Lanswer))
                I12truelist.append(Itar[0][1] / np.linalg.norm(Lanswer))
                I13truelist.append(Itar[0][2] / np.linalg.norm(Lanswer))
                I23truelist.append(Itar[1][2] / np.linalg.norm(Lanswer))
                
                export.append([t[i], I1truelist[-1], I2truelist[-1], I3truelist[-1], I12truelist[-1], I13truelist[-1], I23truelist[-1], omseikai1list[-1], omseikai2list[-1], omseikai3list[-1], Lseikai1list[-1], Lseikai2list[-1], Lseikai3list[-1],
                                     Iest[0][0],  Iest[1][1],  Iest[2][2],  Iest[0][1],  Iest[0][2],  Iest[1][2],  omest[0],  omest[1],  omest[2],  Lest[0],  Lest[1],  Lest[2],
                                     visible_flag, reset, Rest[0], Rest[1], Rest[2], Rest[3], DCM[0], DCM[1], DCM[2]])

                if estcounter/30 >= 0.3 or t[i] < 29:
                    clearConsole()
                    print("t =",t[i],"sec")
                if estcounter == 29:
                    clearConsole()
                    exp = pd.DataFrame(export)
                    exp.explode(0)
                    exp.to_csv(pathres, index=False)
                    plt.pause(0.01)
                    print("t =",t[i]," sec | Results saved as",pathres)
                    estcounter = -1

                if visible_flag == 0 and visible_previous == 0:
                    esttimelist.append(t[i])
                    I1estlist.append(Iest[0][0])
                    I2estlist.append(Iest[1][1])
                    I3estlist.append(Iest[2][2])
                    I12estlist.append(Iest[0][1])
                    I13estlist.append(Iest[0][2])
                    I23estlist.append(Iest[1][2])
                    om1list.append(omest[0])
                    om2list.append(omest[1])
                    om3list.append(omest[2])
                    L1list.append(Lest[0])
                    L2list.append(Lest[1])
                    L3list.append(Lest[2])
                    visible_previous = 0
                if visible_flag == 0 and visible_previous == 1:
                    visible_previous = 0
                if visible_flag == 1:
                    visible_previous = 1

                ax1.cla()
                ax1.scatter(timelist, Lseikai1list, color='#ff0000', marker='.', s=20, label=r'$L_1 true$')
                ax1.scatter(esttimelist, L1list, color='#ffb3b3', marker='.', s=20, label=r'$L_1$ est')
                ax1.scatter(timelist, Lseikai2list, color='#0000ff', marker=',', s=20, label=r'$L_2 true$')
                ax1.scatter(esttimelist, L2list, color='#b3b3ff', marker=',', s=20, label=r'$L_2$ est')
                ax1.scatter(timelist, Lseikai3list, color='#009900', marker='x', s=20, label=r'$L_3$ true')
                ax1.scatter(esttimelist, L3list, color='#4dff4d', marker='x', s=20, label=r'$L_3$ est')

                ax2.cla()
                ax2.scatter(timelist, omseikai1list, color='#ff0000', marker='.', label=r'$\omega_1$ true')
                ax2.scatter(esttimelist, om1list, color='#ffb3b3', marker='.', label=r'$\omega_1$ est')
                ax2.scatter(timelist, omseikai2list, color='#0000ff', marker=',', label=r'$\omega_2$ true')
                ax2.scatter(esttimelist, om2list, color='#b3b3ff', marker=',', label=r'$\omega_2$ est')
                ax2.scatter(timelist, omseikai3list, color='#009900', marker='^', label=r'$\omega_3$ true')
                ax2.scatter(esttimelist, om3list, color='#4dff4d', marker='^', label=r'$\omega_3$ est')

                ax3.cla()
                ax3.scatter(timelist, I1truelist, color='#ff0000', marker='o', label=r'$I_{1}\,true$')
                ax3.scatter(esttimelist, I1estlist, color='#ffb3b3', marker='o', label=r'$I_1$ est')
                ax3.scatter(timelist, I2truelist, color='#0000ff', marker='v', label=r'$I_{2}\,true$')
                ax3.scatter(esttimelist, I2estlist, color='#b3b3ff', marker='v', label=r'$I_2$ est')
                ax3.scatter(timelist, I3truelist, color='#009900', marker='D', label=r'$I_{3}\,true$')
                ax3.scatter(esttimelist, I3estlist, color='#4dff4d', marker='D', label=r'$I_3$ est')
                ax3.scatter(timelist, I12truelist, color='#e6ac00', marker='^', label=r'$I_{12}\,true$')
                ax3.scatter(esttimelist, I12estlist, color='#ffd966', marker='^', label=r'$I_{12}$ est')
                ax3.scatter(timelist, I13truelist, color='#bf00ff', marker='+', label=r'$I_{13}\,true$')
                ax3.scatter(esttimelist, I13estlist, color='#ecb3ff', marker='+', label=r'$I_{13}$ est')
                ax3.scatter(timelist, I23truelist, color='#0099cc', marker='.', label=r'$I_{23}\,true$')
                ax3.scatter(esttimelist, I23estlist, color='#4dd2ff', marker='.', label=r'$I_{23}$ est')

                ax1.legend(bbox_to_anchor=(1, 1.01), loc='lower right', borderaxespad=0, ncol=3, fontsize=12)
                ax1.set_xlabel("Time, sec", fontsize=18)
                ax1.set_ylabel("Angular Momentum", fontsize=18)
                ax2.legend(bbox_to_anchor=(1, 1.01), loc='lower right', borderaxespad=0, ncol=3, fontsize=12)
                ax2.set_xlabel("Time, sec", fontsize=18)
                ax2.set_ylabel("Angular Velocity, rad/s", fontsize=18)
                ax3.legend(bbox_to_anchor=(1.7, 1.0), loc='upper right', borderaxespad=0, ncol=3, fontsize=12)
                ax3.set_xlabel("Time, sec", fontsize=18)
                ax3.set_ylabel("Inertia Tensor", fontsize=18)

                ax1.set_xlim([0, n*dts])
                ax2.set_xlim([0, n*dts])
                ax3.set_xlim([0, n*dts])
                estcounter = estcounter + 1

        counter = counter + 1

    plt.savefig(pathresf)
    exp = pd.DataFrame(export)
    exp.explode(0)
    exp.to_csv(pathres, index=False)
    plt.savefig(pathresf)
    clearConsole()
    print("\n\nSimulation finalized at", int(esttimelist[-1]), "sec | Results saved as", pathres, "and", pathresf,"\n\n")

    return pathnpz, pathtosim, pathtoest, pathres

def Heikojiku(vec):
    ax, ay, az = vec[0], vec[1], vec[2]
    return np.array([[ay ** 2 + az ** 2, -ay * ax, -az * ax], [-ax * ay, ax ** 2 + ay ** 2, -az * ay],
                     [-ax * az, -ay * az, ax ** 2 + ay ** 2]])

# Receives variables from GUI and calls pnp function
def MS(e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e21, e22, e23, e24, e25, e26, e27, e28, e29, e30, e31, e32, e33, e34, e35, e36, e37, e38):

    tic = time.time()

    surfaces = int(e27.get())

    if surfaces != 2 and surfaces != 1:
        print("Number of surfaces should be 1 or 2, but the input value was",surfaces)
        print("Changing the number of surfaces to 2")
        time.sleep(5)

    dts = float(e28.get())
    dte = float(e29.get())
    B   = float(e30.get())
    ST  = float(e31.get())
    IL  = float(e32.get())

    n = int(float(e11.get()) / dts)

    M = float(e8.get())         # SC body mass

    Msap = float(e9.get())      # Solar array panel (SAP) mass

    f = float(e10.get())        # Sensor pixel size in px

    # a,b,c,d,e,g are dimensions of the target body
    a, b, c = float(e1.get()) / 2, float(e2.get()) / 2, float(e3.get()) / 2
    d, e, g = float(e4.get()) / 2, float(e5.get()) / 2, float(e6.get()) / 2

    # SAP Vertices
    SAP1 = np.array(
        [[g, 0, 0, 0], [g, 2*e, 0, 0], [g, 2*e, -2*d, 0], [g, 0, -2*d, 0], [-g, 0, 0, 0], [-g, 2*e, 0, 0], [-g, 2*e, -2*d, 0],
         [-g, 0, -2*d, 0],[0,e,-d,0]])
    SAP2 = np.array(
        [[g, 0, 0, 0], [g, -2*e, 0, 0], [g, -2*e, -2*d, 0], [g, 0, -2*d, 0], [-g, 0, 0, 0], [-g, -2*e, 0, 0], [-g, -2*e, -2*d, 0],
         [-g, 0, -2*d, 0],[0,-e,-d,0]])

    # SAP Rotation
    phy1 = float(e7.get()) * np.pi / 180
    R1 = np.array([[1, 0, 0, 0], [0, cos(phy1), -sin(phy1), 0], [0, sin(phy1), cos(phy1), 0], [0, 0, 0, 1]])
    R2 = np.array([[1, 0, 0, 0], [0, cos(-phy1), -sin(-phy1), 0], [0, sin(-phy1), cos(-phy1), 0], [0, 0, 0, 1]])
    R1e = np.array([[1, 0, 0], [0, cos(phy1), -sin(phy1)], [0, sin(phy1), cos(phy1)]])
    R2e = np.array([[1, 0, 0], [0, cos(-phy1), -sin(-phy1)], [0, sin(-phy1), cos(-phy1)]])
    for i in range(9):
        SAP1[i] = np.dot(R1, SAP1[i].T).T + np.array([0, b,  c, 0])
        SAP2[i] = np.dot(R2, SAP2[i].T).T + np.array([0, -b, c, 0])

    # Inertia Tensor
    Ixx = 1 / 3 * M * (b ** 2 + c ** 2)
    Iyy = 1 / 3 * M * (c ** 2 + a ** 2)
    Izz = 1 / 3 * M * (a ** 2 + b ** 2)
    Ixyz = np.array([[Ixx, 0, 0], [0, Iyy, 0], [0, 0, Izz]])
    Isap = Msap * np.array(
        [[1 / 3 * (e ** 2 + d ** 2), 0, 0], [0, 1 / 3 * (g ** 2 + d ** 2), 0], [0, 0, 1 / 3 * (g ** 2 + e ** 2)]])
        
    Cg1 = ((SAP1[8] * Msap + SAP2[8] * Msap) / (M + 2*Msap))[:3]
    
    Iall = (Ixyz + M * Heikojiku(Cg1)) + (
            np.dot(np.dot(R1e, Isap), R1e.T) + Msap * Heikojiku(SAP1[8,:3] - Cg1)) + (
                   np.dot(np.dot(R2e, Isap), R2e.T) + Msap * Heikojiku(SAP2[8,:3] - Cg1))
    # Observer position
    Xg, Yg, Zg = float(e33.get()), float(e34.get()), float(e35.get())
    vx, vy, vz = float(e36.get()), float(e37.get()), float(e38.get())
    
    cube = np.array([[a, -b, c, 1],[a, b, c, 1],[a, b, -c, 1],[a, -b, -c, 1],[-a, -b, c, 1],[-a, b, c, 1],[-a, b, -c, 1],[-a, -b, -c, 1], \
                     SAP1[0],SAP1[1],SAP1[2],SAP1[3],SAP1[4],SAP1[5],SAP1[6],SAP1[7],
                     SAP2[0],SAP2[1],SAP2[2],SAP2[3],SAP2[4],SAP2[5],SAP2[6],SAP2[7],
                     [a, -b, -c, 1],[a, b, -c, 1],[-a, -b, -c, 1],[-a, b, -c, 1],
                     [a / 2, b / 2, -c, 1], [a / 2, 0, -c, 1], [0, b / 2, -c, 1], [a, 0, -c / 2, 1],    # Markers
                     [0, 0, -1, 1], [1, 0, 0, 1]])
    if surfaces == 1:
        cube[31] = [a, 0, -c, 1]

    mrk     = [cube[28][0],cube[28][1],cube[28][2],cube[29][0],cube[29][1],cube[29][2],cube[30][0],cube[30][1],cube[30][2],cube[31][0],cube[31][1],cube[31][2]]
    markers = np.array([mrk])

    omega1, omega2, omega3 = float(e24.get()), float(e25.get()), float(e26.get())

    phi1, phi2, phi3 = float(e21.get()), float(e22.get()), float(e23.get())

    # --------------------計算条件---------------------
    # dt:積分タイムステップ[s]
    # para_error: 観測誤差[pix]
    # changetime: 片翼が取れるタイムステップ

    para_error = 0  # 1.010#1.01#0.0001 #0.0001

    mt.destroy() # Closes the GUI

    tlist = []
    marker1list, marker2list, marker3list, marker4list = [], [], [], []

    if __name__ == '__main__':

        pathcd = os.path.dirname(__file__)
        pathdata = os.path.join(pathcd, 'data')
        pathtoest = os.path.join(pathdata, 'toestimator.csv')
        pathtosim = os.path.join(pathdata, 'tosimulator.csv')
        pathshape = os.path.join(pathdata,        'shape.csv')
        pathresfo = os.path.join(pathcd, 'results')
        pathres = os.path.join(pathresfo, 'results.csv')
        pathresf = os.path.join(pathresfo, 'results.png')
        pathexef = os.path.join(pathcd, 'ME')

        if surfaces == 1:
            pathexef = os.path.join(pathexef, '1')
        else:
            pathexef = os.path.join(pathexef, '2')
        if sys.platform == "darwin":
            pathexef = os.path.join(pathexef, 'macOS')
            pathexe = os.path.join(pathexef, 'ME')
        else:
            pathexef = os.path.join(pathexef, 'Windows')
            pathexe = os.path.join(pathexef, 'ME.exe')

        pathnpz = os.path.join(pathexef, 'estdata.npz')

        paths = [pathcd, pathexe, pathtoest, pathtosim, pathshape, pathres, pathresf, pathnpz, pathdata, pathresfo]

        np.savetxt(pathshape, markers, delimiter=',')

        result = pnp(cube, omega1, omega2, omega3, Xg, Yg, Zg, phi1, phi2, phi3, vx, vy, vz, dts, dte, B, ST, IL, marker1list, marker2list, marker3list, marker4list, para_error, tlist,f, n, Iall, surfaces, paths)
        
        if os.path.exists(result[-2]):
            os.remove(result[-2])
        if os.path.exists(result[-3]):
            os.remove(result[-3])
        if os.path.exists(result[-4]):
            os.remove(result[-4])
        toc = time.time() - tic
        print("Elapsed time is",round(toc, 4),"seconds.\n\n")
        exit()
# GUI

filepath = os.path.dirname(__file__)

mt = tk.Tk()

image = Image.open(os.path.join(filepath,'figures','MESCs.png'))
photo = ImageTk.PhotoImage(image)

image2 = Image.open(os.path.join(filepath,'figures','MeasurementsAngles2.png'))
photo2 = ImageTk.PhotoImage(image2)

photo3 = ImageTk.PhotoImage(Image.open(os.path.join(filepath,'figures','Picture1.png')))

mt.iconphoto(False, photo3)
mt.title('Target Spacecraft Configuration')
mt.geometry("1375x950")
mt.columnconfigure(0, weight=1)
mt.rowconfigure(0, weight=1)

main_frame = tk.Frame(mt)
main_frame.columnconfigure(0, weight=1)
main_frame.rowconfigure(0, weight=1)
main_frame.grid(sticky=tk.N+tk.S+tk.E+tk.W)

my_canvas = tk.Canvas(main_frame)
my_canvas.columnconfigure(0, weight=1)
my_canvas.rowconfigure(0, weight=1)
my_canvas.grid(sticky=tk.N+tk.S+tk.E+tk.W)

my_scrollbar = tk.Scrollbar(main_frame, orient=tk.VERTICAL, command=my_canvas.yview)
my_scrollbar.grid(column=16, row=0, rowspan=14, sticky='NS')

my_canvas.configure(yscrollcommand=my_scrollbar.set)
my_canvas.bind('<Configure>', lambda e: my_canvas.configure(scrollregion=my_canvas.bbox("all")))

master = tk.Frame(my_canvas)
master.columnconfigure(0, weight=1)
master.rowconfigure(0, weight=1)
master.grid(sticky=tk.N+tk.S+tk.E+tk.W)
my_canvas.create_window((0, 0), window=master, anchor='nw')

tk.Label(master, text="Input the Dimensions (refer to image)\nin meters", justify='left').grid(row=0, column=0,
                                                                                               columnspan=2, padx=10,
                                                                                               pady=10, sticky="w")

tk.Label(master, text="a:").grid(row=1, column=0, padx=10, pady=10, sticky="w")
tk.Label(master, text="b:").grid(row=2, column=0, padx=10, pady=10, sticky="w")
tk.Label(master, text="c:").grid(row=3, column=0, padx=10, pady=10, sticky="w")
tk.Label(master, text="d:").grid(row=4, column=0, padx=10, pady=10, sticky="w")
tk.Label(master, text="e:").grid(row=5, column=0, padx=10, pady=10, sticky="w")
tk.Label(master, text="f:").grid(row=6, column=0, padx=10, pady=10, sticky="w")

tk.Label(master, text="Input Solar Array Panel (SAP) position\nin degrees", justify='left').grid(row=7, column=0,
                                                                                                 columnspan=2, padx=10,
                                                                                                 pady=(30, 10),
                                                                                                 sticky="w")

tk.Label(master, text="\u03F4:").grid(row=8, column=0, padx=10, pady=10, sticky="w")

tk.Label(master, text="Input the mass of body and SAPs\nin kg", justify='left').grid(row=9, column=0, columnspan=2,
                                                                                     padx=10, pady=(30, 10), sticky="w")

tk.Label(master, text="Body:").grid(row=10, column=0, padx=10, pady=10, sticky="w")
tk.Label(master, text="Each SAP:").grid(row=11, column=0, padx=10, pady=10, sticky="w")

tk.Label(master, text="Input the focal length of the camera\nin pixels", justify='left').grid(row=12, column=0,
                                                                                              columnspan=2, padx=10,
                                                                                              pady=(30, 10), sticky="w")

tk.Label(master, text="l:").grid(row=13, column=0, padx=10, pady=10, sticky="w")

tk.Label(master, text="Input Simulation Time in seconds").grid(row=14, column=0, columnspan=2, padx=10, pady=(30, 10),
                                                               sticky="w")

tk.Label(master, text="T:").grid(row=15, column=0, padx=10, pady=10, sticky="w")

wid = 12

e1 = tk.Entry(master, width=wid)
e2 = tk.Entry(master, width=wid)
e3 = tk.Entry(master, width=wid)
e4 = tk.Entry(master, width=wid)
e5 = tk.Entry(master, width=wid)
e6 = tk.Entry(master, width=wid)
e7 = tk.Entry(master, width=wid)
e8 = tk.Entry(master, width=wid)
e9 = tk.Entry(master, width=wid)
e10 = tk.Entry(master, width=wid)
e11 = tk.Entry(master, width=wid)

e1.grid(row=1, column=1, padx=10, pady=10);
e1.insert(0, "6.0")
e2.grid(row=2, column=1, padx=10, pady=10);
e2.insert(0, "4.0")
e3.grid(row=3, column=1, padx=10, pady=10);
e3.insert(0, "4.0")
e4.grid(row=4, column=1, padx=10, pady=10);
e4.insert(0, "1.0")
e5.grid(row=5, column=1, padx=10, pady=10);
e5.insert(0, "12.0")
e6.grid(row=6, column=1, padx=10, pady=10);
e6.insert(0, "5.0")
e7.grid(row=8, column=1, padx=10, pady=10);
e7.insert(0, "30.0")
e8.grid(row=10, column=1, padx=10, pady=10);
e8.insert(0, "500.0")
e9.grid(row=11, column=1, padx=10, pady=10);
e9.insert(0, "30.0")
e10.grid(row=13, column=1, padx=10, pady=10);
e10.insert(0, "2000.0")
e11.grid(row=15, column=1, padx=10, pady=10);
e11.insert(0, "2139")

tk.Label(master, text="Input the Initial Conditions, including orientation and position of the target\n", justify='left').grid(row=0, column=2, columnspan=10, padx=10, pady=10, sticky="w")

tk.Label(master, text="Input time step information\nfor simulator and estimator", justify='left').grid(row=7, column=2, columnspan=7, padx=10, pady=10, sticky="w")

SUB = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
SUP = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
"H2SO4".translate(SUB)

tk.Label(master, text="\u03C61:".translate(SUB)).grid(row=1, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="rad").grid(row=1, column=6, columnspan=3, padx=10, pady=10, sticky="w")
tk.Label(master, text="\u03C62:".translate(SUB)).grid(row=2, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="rad").grid(row=2, column=6, columnspan=3, padx=10, pady=10, sticky="w")
tk.Label(master, text="\u03C63:".translate(SUB)).grid(row=3, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="rad").grid(row=3, column=6, columnspan=3, padx=10, pady=10, sticky="w")
tk.Label(master, text="\u03C91:".translate(SUB)).grid(row=4, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="rad/sec").grid(row=4, column=6, columnspan=3, padx=10, pady=10, sticky="w")
tk.Label(master, text="\u03C92:".translate(SUB)).grid(row=5, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="rad/sec").grid(row=5, column=6, columnspan=3, padx=10, pady=10, sticky="w")
tk.Label(master, text="\u03C93:".translate(SUB)).grid(row=6, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="rad/sec").grid(row=6, column=6, columnspan=3, padx=10, pady=10, sticky="w")

tk.Label(master, text="Sim.:".translate(SUB)).grid(row=8, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="sec").grid(row=8, column=6, columnspan=3, padx=10, pady=10, sticky="w")
tk.Label(master, text="Est.:".translate(SUB)).grid(row=9, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="sec").grid(row=9, column=6, columnspan=3, padx=10, pady=10, sticky="w")

wid2 = 12

e21 = tk.Entry(master, width=wid2)
e22 = tk.Entry(master, width=wid2)
e23 = tk.Entry(master, width=wid2)
e24 = tk.Entry(master, width=wid2)
e25 = tk.Entry(master, width=wid2)
e26 = tk.Entry(master, width=wid2)
e27 = tk.Entry(master, width=2)
e28 = tk.Entry(master, width=wid2)
e29 = tk.Entry(master, width=wid2)

e21.grid(row=1, column=4, columnspan=2, padx=10, pady=10);
e21.insert(0, "3.14159265359")
e22.grid(row=2, column=4, columnspan=2, padx=10, pady=10);
e22.insert(0, "0.0")
e23.grid(row=3, column=4, columnspan=2, padx=10, pady=10);
e23.insert(0, "0.0")
e24.grid(row=4, column=4, columnspan=2, padx=10, pady=10);
e24.insert(0, "0.02")
e25.grid(row=5, column=4, columnspan=2, padx=10, pady=10);
e25.insert(0, "0.12")
e26.grid(row=6, column=4, columnspan=2, padx=10, pady=10);
e26.insert(0, "0.12")
e28.grid(row=8, column=4, columnspan=2, padx=10, pady=10);
e28.insert(0, "0.1")
e29.grid(row=9, column=4, columnspan=2, padx=10, pady=10);
e29.insert(0, "1")

img_label = tk.Label(master, image=photo)
img_label.image = photo
img_label.grid(row=1, column=12, columnspan=7, rowspan=8, padx=10)

img_label2 = tk.Label(master, image=photo2)
img_label2.image2 = photo2
img_label2.grid(row=8, column=12, columnspan=7, rowspan=8, padx=10)
tk.Label(master, text="Assuming that origin is in center of the main body, markers are in the positions:\n(1.5, 1, -2), (1.5, 0, -2), (0, 1, -2), (3, 0, -1), in case of 2 planes\n(1.5, 1, -2), (1.5, 0, -2), (0, 1, -2), (3, 0, -2), in case of 1 plane", justify='left').grid(row=0, column=12, columnspan=7, padx=10, pady=10, sticky="w")

tk.Label(master, text="Amount of past estimations used\nby the estimator (bandwidth):", justify='left').grid(row=10, column=2, columnspan=7, padx=10, pady=10, sticky="w")

tk.Label(master, text="B:").grid(row=11, column=2, columnspan=2, padx=10, pady=10, sticky="w");
tk.Label(master, text="steps").grid(row=11, column=6, columnspan=3, padx=10, pady=10, sticky="w")

e30 = tk.Entry(master, width=wid2)
e30.grid(row=11, column=4, columnspan=2, padx=10, pady=10)
e30.insert(0, "1200")

tk.Label(master, text="Amount data stored to start\nestimation:", justify='left').grid(row=12, column=2, columnspan=7, padx=10, pady=10, sticky="w")

tk.Label(master, text="ST:").grid(row=13, column=2, columnspan=2, padx=10, pady=10, sticky="w")
tk.Label(master, text="steps").grid(row=13, column=6, columnspan=3, padx=10, pady=10, sticky="w")

e31 = tk.Entry(master, width=wid2)
e31.grid(row=13, column=4, columnspan=2, padx=10, pady=10)
e31.insert(0, "0")

tk.Label(master, text="Number of times \"I\" is stored\nfor each update of \"L\":", justify='left').grid(row=14, column=2, columnspan=7, padx=10, pady=10, sticky="w")

tk.Label(master, text="IL:").grid(row=15, column=2, columnspan=2, padx=10, pady=10, sticky="w")
tk.Label(master, text="steps").grid(row=15, column=6, columnspan=3, padx=10, pady=10, sticky="w")

e32 = tk.Entry(master, width=wid2)
e32.grid(row=15, column=4, columnspan=2, padx=10, pady=10)
e32.insert(0, "1")

tk.Label(master, text="R1:".translate(SUB)).grid(row=1, column=9, columnspan=1, padx=10, pady=10, sticky="w");
tk.Label(master, text="m").grid(row=1, column=11, columnspan=1, padx=10, pady=10, sticky="w")
tk.Label(master, text="R2:".translate(SUB)).grid(row=2, column=9, columnspan=1, padx=10, pady=10, sticky="w");
tk.Label(master, text="m").grid(row=2, column=11, columnspan=1, padx=10, pady=10, sticky="w")
tk.Label(master, text="R3:".translate(SUB)).grid(row=3, column=9, columnspan=1, padx=10, pady=10, sticky="w");
tk.Label(master, text="m").grid(row=3, column=11, columnspan=1, padx=10, pady=10, sticky="w")
tk.Label(master, text="V1:".translate(SUB)).grid(row=4, column=9, columnspan=1, padx=10, pady=10, sticky="w");
tk.Label(master, text="m/sec").grid(row=4, column=11, columnspan=1, padx=10, pady=10, sticky="w")
tk.Label(master, text="V2:".translate(SUB)).grid(row=5, column=9, columnspan=1, padx=10, pady=10, sticky="w");
tk.Label(master, text="m/sec").grid(row=5, column=11, columnspan=1, padx=10, pady=10, sticky="w")
tk.Label(master, text="V3:".translate(SUB)).grid(row=6, column=9, columnspan=1, padx=10, pady=10, sticky="w");
tk.Label(master, text="m/sec").grid(row=6, column=11, columnspan=1, padx=10, pady=10, sticky="w")

e33 = tk.Entry(master, width=wid2)
e34 = tk.Entry(master, width=wid2)
e35 = tk.Entry(master, width=wid2)
e36 = tk.Entry(master, width=wid2)
e37 = tk.Entry(master, width=wid2)
e38 = tk.Entry(master, width=wid2)

e33.grid(row=1, column=10, columnspan=1, padx=10, pady=10);
e33.insert(0, "0.0")
e34.grid(row=2, column=10, columnspan=1, padx=10, pady=10);
e34.insert(0, "0.0")
e35.grid(row=3, column=10, columnspan=1, padx=10, pady=10);
e35.insert(0, "16")
e36.grid(row=4, column=10, columnspan=1, padx=10, pady=10);
e36.insert(0, "0.0")
e37.grid(row=5, column=10, columnspan=1, padx=10, pady=10);
e37.insert(0, "0.0")
e38.grid(row=6, column=10, columnspan=1, padx=10, pady=10);
e38.insert(0, "0.0")

tk.Label(master, text="Markers on how many planes?\n(1 or 2)", justify='left').grid(row=7, column=9, columnspan=2, padx=10, pady=10, sticky="w")

e27.grid(row=7, column=11, columnspan=1, padx=10, pady=10);
e27.insert(0, "2")

if int(e27.get()) != 2 or int(e27.get()) != 1:
    e27.insert(0, "")

tk.Button(master, text='Accept and Simulate', justify='center',
          command=lambda:MS(e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e21, e22, e23, e24, e25, e26, e27, e28, e29, e30, e31, e32, e33, e34, e35, e36, e37, e38)).grid(row=8, column=9, columnspan=3, pady=10)

tk.Label(master, text="To see more information in the window, scroll\ndown with the mouse cursor over the scrollbar", justify='left').grid(row=9, column=9, columnspan=3, padx=10, pady=10, sticky="w")
tk.Label(master, text="If the background is too dark and the figures'\nannotations are hard to see, we recommend\nturning off system dark mode.", justify='left').grid(row=10, column=9, columnspan=3, rowspan=2, padx=10, pady=10, sticky="w")


tk.mainloop()
