%% DESCRIPTION ------------------------------------------------------------

% This is a sample code to demonstrate how to call the Motion Estimator exe
% cutable
% 
% This example uses a motionless model (only markers positions are output)
    
clear all   % Clear Workspace
close all   % Close All Windows
clc         % Clear Command Window

tic         % Start counting clock

%% SEQUENCE TO IDENTIFY IF IS MATLAB OR OCTAVE ----------------------------

isMATLAB = false;
LIC = license('inuse');
for elem = 1:numel(LIC)
    envStr = LIC(elem).feature;
    if strcmpi(envStr,'matlab')
        isMATLAB = true;
        break
    end
end


%% MARKERS CONFIGURATION --------------------------------------------------

mr  =   [ 1.5  1.5 -1.5;
         -1.5 -1.5 -1.5;
         -1.5  1.5 -1.5;
          1.5 -1.5 -1.5];

mrk =   [mr(1,:)-mr(1,:) mr(2,:)-mr(1,:) mr(3,:)-mr(1,:) mr(4,:)-mr(1,:)];


%% INITIAL MARKERS' POSITIONS ON 2000 X 2000 FRAME ------------------------

M1  =   [ 209.40  686.61];      % Marker 1 X and Y coordinates
M2  =   [-162.39 -556.98];      % ...
M3  =   [ 545.58 -115.38];      % ...
M4  =   [-609.69  133.91];      % ...
M   =   [M1 M2 M3 M4];          % All Markers

                                                                            % Generating figure of above coordinates:

                                                                            motion = figure; ax = axes; axis equal; set (ax,'XLim',[-1000 1000],'YLim',[-1000 1000])
                                                                            title({[""];["Markers' Projections"];['']}); xlabel("X, pixels"); ylabel("Y, pixels")
                                                                            grid on; hold on;
                                                                            
                                                                            l13 = plot([M1(1) M3(1)],[M1(2) M3(2)],'--k');
                                                                            l14 = plot([M1(1) M4(1)],[M1(2) M4(2)],'--k');
                                                                            l23 = plot([M2(1) M3(1)],[M2(2) M3(2)],'--k');
                                                                            l24 = plot([M2(1) M4(1)],[M2(2) M4(2)],'--k');
                                                                            if isMATLAB
                                                                                p1  = plot(M1(1), M1(2),'o','MarkerFaceColor','#0072BD','MarkerEdgeColor','#0072BD');
                                                                                p2  = plot(M2(1), M2(2),'o','MarkerFaceColor','#D95319','MarkerEdgeColor','#D95319');
                                                                                p3  = plot(M3(1), M3(2),'o','MarkerFaceColor','#EDB120','MarkerEdgeColor','#EDB120');
                                                                                p4  = plot(M4(1), M4(2),'o','MarkerFaceColor','#7E2F8E','MarkerEdgeColor','#7E2F8E');
                                                                            else
                                                                                p1  = plot(M1(1), M1(2),'o','MarkerFaceColor',[0.0000 0.4470 0.7410],'MarkerEdgeColor',[0.0000 0.4470 0.7410]);
                                                                                p2  = plot(M2(1), M2(2),'o','MarkerFaceColor',[0.8500 0.3250 0.0980],'MarkerEdgeColor',[0.8500 0.3250 0.0980]);
                                                                                p3  = plot(M3(1), M3(2),'o','MarkerFaceColor',[0.9290 0.6940 0.1250],'MarkerEdgeColor',[0.9290 0.6940 0.1250]);
                                                                                p4  = plot(M4(1), M4(2),'o','MarkerFaceColor',[0.4940 0.1840 0.5560],'MarkerEdgeColor',[0.4940 0.1840 0.5560]);
                                                                            end
                                                                            
                                                                            pause(0.01)


%% ESTIMATORS PARAMETERS --------------------------------------------------

dte =   1;                      % Estimator Interval                [s]

B   =   10;                     % Estimator stored information      [s]

ST  =   0;                      % Steps stored by estimator before
                                % first estimation starts           [s]

IL  = 1;                        % Number of times "I" is estimated 
                                % before "L" updates                [s]
                                
                                
%% PATHS ------------------------------------------------------------------

% Paths required for the files to be exchanged
% between simulator and estimator

cdc     = pwd;                                              % Directory

if ismac
    ptME    = fullfile(cdc,'ME','1','macOS','ME');          % Motion Estmtr
    ptNP    = fullfile(cdc,'ME','1','macOS','estdata.npz'); % Internal file
else
    ptME    = fullfile(cdc,"ME","1","Windows","ME.exe");
    ptNP    = fullfile(cdc,"ME","1","Windows","estdata.npz");
end

ptTE    = fullfile(cdc,'data','toestimator.csv');       % Input  csv file
ptTS    = fullfile(cdc,'data','tosimulator.csv');       % Output csv file
ptSP    = fullfile(cdc,'data',      'shape.csv');       % Marker csv file

ptDT    = fullfile(cdc,'data');                         % Directory for I/O

csvwrite(ptSP,mrk);


%% CALLING ESTIMATOR ------------------------------------------------------

% 1) Input file must be created

t       = 1;                                        % Time information
reset   = 1;                                        % Reset internal file

input   = [M, IL, ST, t, dte, B, reset];            % Input array

csvwrite(ptTE,input);                               % Inpit array to csv

% 2) Motion Esitmator is called

if ismac
    command = sprintf('"%s" "%s" "%s" "%s" "%s"', ptME, ptTE, cdc, ptTS, ptSP);
    system(command);
else
    command = sprintf('"%s" "%s" "%s" "%s" "%s"', ptME, ptTE, cdc, ptTS, ptSP);
    system(command);
end

%% MOTION ESTIMATOR MUST BE CALLED AT LEAST TWICE -------------------------

t       = t + dte;                                  % Time information
reset   = 0;                                        % Keep internal file

input   = [M, IL, ST, t, dte, B, reset];            % Input array

csvwrite(ptTE,input);                               % Inpit array to csv

if ismac
    % Command string to call system
    command = sprintf('"%s" "%s" "%s" "%s" "%s"', ptME, ptTE, cdc, ptTS, ptSP);
    system(command);    % Call executable with inout file and arguments
else
    command = sprintf('"%s" "%s" "%s" "%s" "%s"', ptME, ptTE, cdc, ptTS, ptSP);
    system(command);
end


%% READING OUTPUT FILE ----------------------------------------------------

rslt    =   csvread(ptTS);                          % Read output csv file

disp('Markers Positions [in meters, w.r.t. observer]:');
M1r     = rslt(6,:); disp(['Marker 1: [',num2str(M1r),']']);
M2r     = rslt(7,:); disp(['Marker 2: [',num2str(M2r),']']);
M3r     = rslt(8,:); disp(['Marker 3: [',num2str(M3r),']']);
M4r     = rslt(9,:); disp(['Marker 4: [',num2str(M4r),']']); disp(' ');

DCM     = rslt(10:12,:); disp('Direction Cosine Matrix:')
                         disp(['[',num2str(DCM(1,:))]);
                         disp([' ',num2str(DCM(2,:))]);
                         disp([' ',num2str(DCM(3,:)),']']); disp(' ');

disp('Estimated distances between markers [in meters]:');
D12     = norm(M1r-M2r); disp(['D12: ',num2str(D12)]);
D13     = norm(M1r-M3r); disp(['D13: ',num2str(D13)]);
D14     = norm(M1r-M4r); disp(['D14: ',num2str(D14)]);
D23     = norm(M2r-M3r); disp(['D23: ',num2str(D23)]);
D24     = norm(M2r-M4r); disp(['D24: ',num2str(D24)]);
D34     = norm(M3r-M4r); disp(['D34: ',num2str(D34)]); disp(' ');


